; Game Raccoon controller_state.68k.asm
; Game Raccoon Revision 1 68000 Menu Firmware
; (c) Mathew Carr July 2020
;
; Reads controller input from the two front controller ports.
;
; Copyright (c) 2020 Mathew Carr
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU Affero General Public License Version 3
; with Additional Terms as found in the file
; LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU Affero General Public License for more details.
; 
; You should have received a copy of the GNU Affero General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.

; To initialise the controller status struct:
; 
;        lea     pad_state,a0                                                ; Set a0 to a struct of size PAD_STATUS_TOTAL_SIZE in your memory.
;        jsr     controllers_reset                                           ; Initialise the controller status struct.
;
; To update the controller status struct every frame:
;
;        lea     pad_state,a0                                                ; Set a0 to a struct of size PAD_STATUS_TOTAL_SIZE in your memory.
;        jsr     controllers_read_and_derive                                 ; Populate the struct with the current controller state
;                                                                            ; and generate frame-to-frame change flags.
;
; To test for button press:
;
;        btst    #CONTROLLER_S_BIT_btst_l,pad_state+PAD_1+CURR+LOW_BYTE      ; Z flag is ZERO if Start is CURRently depressed.
;        bne     button_not_pressed                                          ; BNE branch is taken if button is idle.
;        ; button pressed response
;        jmp     done
; button_not_pressed:       
;        ; button idle response
;        jmp     done
;
; Change CURR to PREV to see previous state, or PRESSED, HELD, RELEASED or IDLE to query
; frame-to-frame button state changes.
;
; Be aware that buttons MZYXP6 are in the High byte of the two-byte word where
; the flags are stored. Since btst can only target the eight bits of a byte
; when you are targeting a constant address in memory directly,
; you query these upper buttons using the following:
;        btst    #CONTROLLER_M_BIT_btst_h,pad_state+PAD_1+CURR+LOW_BYTE      ; Z flag is ZERO if Mode is CURRently depressed.
;
; If you have the flag value stored in a register, you can use the CONTROLLER_*_BIT
; constants directly.

; --------------------------

; Constants for controller state management:

               ;FEDCBA9876543210
CONTROLLER_U = %0000000000000001
CONTROLLER_D = %0000000000000010
CONTROLLER_L = %0000000000000100
CONTROLLER_R = %0000000000001000
CONTROLLER_B = %0000000000010000
CONTROLLER_C = %0000000000100000
CONTROLLER_A = %0000000001000000
CONTROLLER_S = %0000000010000000
CONTROLLER_M = %0000000100000000
CONTROLLER_Z = %0000001000000000
CONTROLLER_Y = %0000010000000000
CONTROLLER_X = %0000100000000000
CONTROLLER_P = %0100000000000000
CONTROLLER_6 = %1000000000000000

; Active LOW!
; Use these for BTST if we've loaded a controller memory word into a register first.
CONTROLLER_U_BIT = 0
CONTROLLER_D_BIT = 1
CONTROLLER_L_BIT = 2
CONTROLLER_R_BIT = 3
CONTROLLER_B_BIT = 4
CONTROLLER_C_BIT = 5
CONTROLLER_A_BIT = 6
CONTROLLER_S_BIT = 7
CONTROLLER_M_BIT = 8  ; Mode.
CONTROLLER_Z_BIT = 9
CONTROLLER_Y_BIT = 10
CONTROLLER_X_BIT = 11
CONTROLLER_P_BIT = 14 ; Depressed if a controller is present at all.
CONTROLLER_6_BIT = 15 ; Depressed we if we are using a six button pad. :3

CONTROLLER_U_BIT_btst_l = 0 ; BTST only works on bytes, not words, when hitting a memory location directly for data source.
CONTROLLER_D_BIT_btst_l = 1 ; We need to target the low byte (+1) if we're BTSTing for this!
CONTROLLER_L_BIT_btst_l = 2
CONTROLLER_R_BIT_btst_l = 3
CONTROLLER_B_BIT_btst_l = 4
CONTROLLER_C_BIT_btst_l = 5
CONTROLLER_A_BIT_btst_l = 6
CONTROLLER_S_BIT_btst_l = 7
CONTROLLER_M_BIT_btst_h = 0 ; We need to target the high byte (+0) if we're BTSTing for this!
CONTROLLER_Z_BIT_btst_h = 1
CONTROLLER_Y_BIT_btst_h = 2
CONTROLLER_X_BIT_btst_h = 3
CONTROLLER_P_BIT_btst_h = 6
CONTROLLER_6_BIT_btst_h = 7     

; Define the controller state struct as a series of word offsets:

; The controller reading function takes as a parameter
; the address of the struct to write to, so these are not
; absolute addresses.

; These are true = zero values.

PAD_1                    = 0

CURR                     = (2*0)
PREV                     = (2*1)

PRESSED                  = (2*2)       ; Pressed  =  curr & !prev
HELD                     = (2*3)       ; Held     =  curr &  prev
RELEASED                 = (2*4)       ; Released = !curr &  prev
IDLE                     = (2*5)       ; Idle     = !curr & !prev

PAD_STATUS_STRUCT_STRIDE = (2*6)

PAD_2                    = PAD_STATUS_STRUCT_STRIDE

PAD_STATUS_TOTAL_SIZE    = (2*PAD_STATUS_STRUCT_STRIDE)

HIGH_BYTE                = 0
LOW_BYTE                 = 1

; ---------------------------
; CONTROLLER READING ROUTINES
; ---------------------------

; These routines read the controllers and populate
; a PAD_STATUS_TOTAL_SIZE region of memory with
; the current and previous state of the pads.
; These flags are 1 idle, 0 pressed (or 'true'
; for derived states).

; Read the controllers and populate the controller status
; struct pointed to by a0.
; Populates curr, prev and all derived states.
; Invalidates all registers.
; a0 is address of controller status struct.
controllers_read_and_derive:
        movem.l a0,-(a7)
        jsr     controllers_read
        movem.l (a7)+,a0
        jsr     controllers_derive
        rts

; Reset the controllers to known unpressed state.
; Invalidates all registers.
; a0 is address of controller status struct.
controllers_reset:
        move.l  #$FFFFFFFF,PAD_1(a0)                           ; This should clear both low and high words! (Curr and prev are stored consecutively.)
        move.l  #$FFFFFFFF,PAD_2(a0)
        jsr     controllers_derive
        rts

; Read the controllers and populate the controller status
; struct pointed to by a0.
; Only populates curr and prev.
; Invalidates all registers.
; a0 is address of controller status struct.
controllers_read:
        lea     MM_CONTROLLER_1_DATA+1,a1                      ; a1 is the memory mapped register for the data source.
        
        move.w  #1,d3                                          ; Go through two pads.

controllers_read_next_controller:
        move.w  CURR(a0),PREV(a0)                              ; Copy current state to previous state.

        ; We want to read both controller 1 and controller 2. We do this in a loop because the mm registers are close.
        move.b  #$40,MM_CONTROLLER_DATA_TO_CONTROL_OFFSET(a1)  ; Set direction of bits of port 1.

        ; We're going to try and read the controller port in such a way that's compatible with both the 3 button and 6 button pad.
        ; https://github.com/jonthysell/SegaController/wiki/How-To-Read-Sega-Controllers
                                                               ; Enter this with TH low. (Cycle state 0.)
        
        move.b  #$40,(a1)                                      ; Make TH high. (Cycle state 1.)
        nop                                                    ; Allow the controller to respond.
        nop

        move.b  #$00,(a1)                                      ; Make TH low.  (Cycle state 2.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a1),d1                                        ; d1 = xxxxxxxxxxSA!!xx --- exclamation marks are double LOW if a controller is present.
        move.b  d1,d2                                          ; d2 = 0000000000001100
        and.b   #%00001100,d2                                  ; d2 = zero if controller present, else no controller.
        beq     controllers_read_present_continue
        move.w  #%1111111111111111,d0                          ; d0 = 1111111111111111 controller fully idle, not present
        jmp     controllers_read_done                          ; return d0
controllers_read_present_continue:
        move.w  #%1011111111111111,d0                          ; d0 = 1011111111111111 set the controller present bit
        lsl.b   #2,d1                                          ; d1 =         SAxxxxxx
        or.b    #%00111111,d1                                  ; d1 =         Sa011111 --- active LOW.
        move.b  d1,d0                                          ; d0 = 10111111Sa011111

        move.b  #$40,(a1)                                      ; Make TH high. (Cycle state 3.)
        nop                                                    ; Allow the controller to respond.
        nop        
        move.b  (a1),d1                                        ; d1 =         xxCBRLDU
        or.b    #%11000000,d1                                  ; d1 =         11CBRLDU                
        and.b   d1,d0                                          ; d0 = 10111111SACBRLDU --- active LOW.

        move.b  #$00,(a1)                                      ; Make TH low.  (Cycle state 4.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a1),d1                                        ; d1 =         xxxxxx!! --- exclamation marks are double LOW if the controller is 6 button.
        move.b  #%00000011,d2                                  ; d2 =         00000011
        and.b   d1,d2                                          ; d2 = zero if six button controller present, else three button.
        bne     controllers_read_three_button_found
        and.w   #%0111111111111111,d0                          ; d0 = 00111111SACBRLDU 
controllers_read_three_button_found:                           ; d0 = 6P111111SACBRLDU  
        
        move.b  #$40,(a1)                                      ; Make TH high. (Cycle state 5.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a1),d1                                        ; d1 =         xxxxMXYZ
        lsl.w   #8,d1                                          ; d1 = xxxxMXYZxxxxxxxx
        or.w    #%1111000011111111,d1                          ; d1 = 1111MXYZ11111111
        and.w   d1,d0                                          ; d0 = 6P11MXYZSACBRLDU        
        
        move.b  #$00,(a1)                                      ; Make TH low.  (Cycle state 6.)
        nop                                                    ; Allow the controller to respond.
        nop                                                    ; Let the controller idle to reset the counter. Leave TH low.
        
controllers_read_done:
        move.w  d0,(a0)

        lea     MM_CONTROLLER_DATA_SOURCE_STRIDE(a1),a1        ; Go to the next data source memory mapped register location
        lea     PAD_STATUS_STRUCT_STRIDE(a0),a0                ; Go to the next pad status struct
                                                                   
        dbf     d3,controllers_read_next_controller

        rts

; Make the inter-frame comparison values based on the current and previous states of the controllers.
; Invalidates all registers.
; a0 is address of controller status struct.
controllers_derive:
        move.w  #1,d5                                   ; Loop through two controllers.
        
controllers_derive_next_controller:
        move.w  (a0)+,d0                                ; Load current into d0
        move.w  (a0)+,d1                                ; Load previous into d1

        move.w  d0,d2
        not.w   d2                                      ; d2 = !curr
        move.w  d1,d3
        not.w   d3                                      ; d3 = !prev
        ; Remember that the input values are active LOW, so our derived values are active LOW.
        
        move.w  d0,d4
        or.w    d3,d4                                   ; d4 =  curr | !prev = PRESSED
        move.w  d4,(a0)+

        move.w  d0,d4
        or.w    d1,d4                                   ; d4 =  curr |  prev = HELD
        move.w  d4,(a0)+

        move.w  d2,d4
        or.w    d1,d4                                   ; d4 = !curr |  prev = RELEASED
        move.w  d4,(a0)+

        move.w  d2,d4
        or.w    d3,d4                                   ; d4 = !curr | !prev = IDLE
        move.w  d4,(a0)+

        dbf     d5,controllers_derive_next_controller

        rts