# Game Raccoon fruitsalad.py
# (c) Mathew Carr July 2020
#
# The purpose of this file is to automate various inclusion tasks
# to allow imported files to have associated location and size
# symbols, and trigger the generation of static screens such as
# the predefined tegel information screens.
#
# Copyright (c) 2020 Mathew Carr
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License Version 3
# with Additional Terms as found in the file
# LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Usage:
# python fruitsalad.py stage_name
#
# Where stage_name is one of:
# generate_included_binaries_includer:
#    Make included_binaries.68k.asm containing asm directives for
#    including all the indicated files, automatically making
#    symbolic names for useful things.
# compile_text_screens:
#    Compile static text into static blob of tegels.
# compile_entry_point_ramresident:
#    Create the game raccoon boot image file containing the entry point stub,
#    the incbinned binaries, and the incbinned .bin RAM-resident part all together.
#    This is the complete Game Raccoon boot image, except without any directory cache present.
# append_imitation_directory_cache_binary:
#    Create a faux full 4Mbyte boot partition binary by combining the bare Game Raccoon
#    boot ROM image with an imitation directory cache (created by imitate_directory_cache.py).
#    You need this to test the Game Raccoon boot ROM on an emulator since there wouldn't
#    be any directory cache otherwise.
# complete_boot_rom_image:
#    All of the above, resulting in:
#    game_raccoon_boot_image.bin 
#       Game Raccoon 68k Mega Drive menu firmware image.
#    game_raccoon_boot_image_directory_cache.bin 
#       Game Raccoon 68k Mega Drive menu firmware image with appended directory
#       record cache suitable for use in emulators for testing.

import os
import argparse

# Give a list of binaries to include here (resources, etc.),
# and the Python code will generate an included_binaries.68k.asm
# file incbinning all the indicated files together with associated
# location and size symbols.
INCLUDED_BINARIES = {
   "menutexture.img.bin":                                                 "menutexture_img_bin",
   "intro_screen_plane_b_texture_tegels.bin":                             "intro_screen_plane_b_texture_tegels",
   "intro_game_raccoon_logo_image.bin":                                   "intro_game_raccoon_logo_image_bin",
   "sprites_buttons_abcs_image.bin":                                      "sprites_buttons_abcs_image_bin",
   "sprites_raccoon_selection_hand_arrow_widgets.bin":                    "sprites_raccoon_selection_hand_arrow_widgets_bin",
   "about_screen_predefined_tegels.bin":                                  "about_screen_predefined_tegels_bin",
   "flash_warning_screen_predefined_tegels.bin":                          "flash_warning_screen_predefined_tegels_bin",
   # 8x8fonts_grafx2_linux_by_usr_share:
   "saikyo_torus_8x8_font_image.bin":                                     "saikyo_torus_8x8_font_image_bin",
   # tepples gba gsmplayer font:
   "tepples_gsmplay_8x16_font_image.bin":                                 "tepples_gsmplay_8x16_font_image_bin",
}

if __name__ == "__main__":
   parser = argparse.ArgumentParser()
   parser.add_argument("stage")
   args = parser.parse_args()
   stage = args.stage
   
   # fruitsalad stages are like Makefile targets except I say so.
   
   # Make included_binaries.68k.asm containing asm directives for
   # including all the indicated files, automatically making
   # symbolic names for useful things.
   if stage == "generate_included_binaries_includer":
      with open('included_binaries.68k.asm', 'w') as included_binaries_68k_asm_file:
         for ib_filename, symbol_name in INCLUDED_BINARIES.items():
            symbol_name_end = symbol_name + "_end"
         
            included_binaries_68k_asm_file.write("        align   4\n")
            included_binaries_68k_asm_file.write("{symbol}:\n".format(symbol=symbol_name))
            included_binaries_68k_asm_file.write("        incbin  \"{filename}\"\n".format(filename=ib_filename))
            included_binaries_68k_asm_file.write("{symbol} = *\n".format(symbol=symbol_name_end))
            included_binaries_68k_asm_file.write("\n")
            included_binaries_68k_asm_file.write("{symbol_bytes: <50} = ({end: <50}-{beg: <50})\n".format(symbol_bytes=symbol_name+"_length_bytes", end=symbol_name_end, beg=symbol_name))
            included_binaries_68k_asm_file.write("{symbol_words: <50} = (({symbol_bytes: <50}+1)/2)\n".format(symbol_words=symbol_name+"_length_words", symbol_bytes=symbol_name+"_length_bytes"))
            included_binaries_68k_asm_file.write("{symbol_longs: <50} = (({symbol_bytes: <50}+3)/4)\n".format(symbol_longs=symbol_name+"_length_longs", symbol_bytes=symbol_name+"_length_bytes"))
            included_binaries_68k_asm_file.write("\n")
               
      exit(0)
            
   # Compile static text into static blob of tegels.
   if stage == "compile_text_screens":
      os.system('python intro_screen_text_compiler.py ABOUT_SCREEN_LINES about_screen_predefined_tegels.bin')
      os.system('python intro_screen_text_compiler.py FLASH_WARNING_SCREEN_LINES flash_warning_screen_predefined_tegels.bin')
      exit(0)
            
   # Create the game raccoon boot image file containing the entry point stub,
   # the incbinned binaries, and the incbinned .bin RAM-resident part all together.
   # This is the complete Game Raccoon boot image, except without any directory cache present.
   if stage == "compile_entry_point_ramresident":
      os.system('asm68ksik -C 68K -e -w -b 0 -o game_raccoon_boot_image.bin entry_point.68k.asm')
      exit(0)
         
   # Create a faux full 4Mbyte boot partition binary by combining the bare Game Raccoon
   # boot ROM image with an imitation directory cache (created by imitate_directory_cache.py).
   # You need this to test the Game Raccoon boot ROM on an emulator since there wouldn't
   # be any directory cache otherwise.
   # $000000: Mega Drive header
   # $000200: Entry point
   # $yyyyyy: Padding $00 bytes up to $1FFFFF
   # $200000: Directory cache
   # $zzzzzz: Padding $FF bytes up to $3FFFFF
   if stage == "append_imitation_directory_cache_binary":
      b = open("game_raccoon_boot_image.bin", "rb").read()
      b_appended = b + (b'\x00' * (0x200000 - len(b))) + open("directory_cache_emulator.bin", "rb").read()
      b_appended = b_appended + (b'\xFF' * (0x400000 - len(b_appended)))
      open("game_raccoon_boot_image_directory_cache.bin", "wb").write(b_appended)
      exit(0)
   
   # All of the above targets.
   # End result is game_raccoon_boot_image_directory_cache.bin
   if stage == "complete_boot_rom_image":
      stages = [
         "generate_included_binaries_includer",
         "compile_text_screens",
         "compile_entry_point_ramresident",
         "append_imitation_directory_cache_binary"
      ]
      
      for s in stages:
         result = os.system("python fruitsalad.py %s" % s)
         if result != 0:
            raise Exception(result)
            
      exit(0)
   
   print("Unrecognised stage?")
   exit(-1)