# Game Raccoon intro_screen_text_compiler.py
# (c) Mathew Carr July 2020
#
# Takes in the ABOUT_SCREEN_LINES grid of strings
# and outputs the complete binary blob of tegels
# to load into Plane A to show the about screen.
#
# Copyright (c) 2020 Mathew Carr
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License Version 3
# with Additional Terms as found in the file
# LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Usage:
# python intro_screen_text_compiler.py

import argparse
import datetime

# Maximum width = 40, but leave a 1 cell border around the edge!

# These screens will contain a list of 16-bit tegel values x 64 characters wide
# to fill the entire size of the plane.
# Preceding these will be a 16-bit unsigned value indicating the number of lines available.
# Use this to constrain scrolling.

ABOUT_SCREEN_LINES = (
#   [                                      ] 40 chars wide.
   " \x06Game Raccoon Revision 1, Menu 1", # x06 is large text escape upper half.
   " \x07Game Raccoon Revision 1, Menu 1", # x07 is large text escape lower half.
   "",
   " Hardware and Software Design by:",
   "  \x02Mathew Carr",
   "  mrdictionary.net",
   "",
   " \x05 68k assembled", # The 0x05 escape inserts the current date/time.
   "",
   " Assembled using asmx by:",
   "  Bruce Tomlin",
   "",
   " 16x8 GSMplay font by:",
   "  Damian Yerrick pineight.com",
   "",
   " 8x8 Saikyo, Torus fonts by:",
   "  usr_share OpenGameArt",
   "",
   " Special thanks to:",
   "  Cearn,Charles MacDonald,Chilly Willy,",
   "  Consolemad, Kabuto,oerg866,raq, Sik,",
   "  Stef, Steve Snake, Tepples, TmEE,",
   "  VAdaPEGA.",
   "",
   " Greets to:",
   "  Fizzer, GoBusto, Mungojerrie,",
   "  The Sanctum, YakYak, SpritesMind,",
   "  #uplink, #uberducks,",
   "  and all those who follow the Raccoon!",
   "",
   " Game Raccoon is not produced or",
   "  endorsed by Sega.",
) # No line should be bigger than 40 printable characters. (Not including colour escapes.)

FLASH_WARNING_SCREEN_LINES = ( # How very 'Spindizzy Worlds hidden level editor' is this?
   "",
   " Hey there!",
   "",
   " Looks like you want to re-flash",
   " the Game Raccoon's \x02menu firmware\x01.",
   "",
   " This is a dangerous process that can",
   " render your Game Raccoon inoperable",
   " if interrupted or overwritten with",
   " an invalid firmware file.",
   "",
   " To recover your Game Raccoon in an",
   " emergency, it will be necessary to use",
   " a PIC programmer such as a PICKIT 2 to",
   " directly interact with the on-board",
   " processor and make it rewrite a valid",
   " menu firmware to the Flash module's",
   " boot partition.",
   "",
   " Support may not be available.",
   "",
   " No changes have yet been made.",
   "",
   " \x02To abort, turn your console off now.",
   "",
   " If you accept the responsibility and",
   " the risk, press A + B + C to flash.",
) # No line should be bigger than 40 printable characters. (Not including colour escapes.)

screens = {}

screens["ABOUT_SCREEN_LINES"]         = ABOUT_SCREEN_LINES
screens["FLASH_WARNING_SCREEN_LINES"] = FLASH_WARNING_SCREEN_LINES

# Number of exclamation mark tile number loaded into VRAM.
FONT_8X8_BASE_TILE_NUMBER    = 1441
FONT_8X16_BASE_TILE_NUMBER_H = 1253
FONT_8X16_BASE_TILE_NUMBER_L = FONT_8X16_BASE_TILE_NUMBER_H+94

# Given an ASCII charcode and a palette line,
# produce the final tegel value.
# Nonprintable characters become zeroes.
def tegel(charcode, colour, largetextmode):
   tilenumber = 0
   
   if charcode >= 33:
      if largetextmode == 0x00:
         tilenumber = FONT_8X8_BASE_TILE_NUMBER + (charcode-33)
      if largetextmode == 0x06:
         tilenumber = FONT_8X16_BASE_TILE_NUMBER_H + (charcode-33)
      if largetextmode == 0x07:
         tilenumber = FONT_8X16_BASE_TILE_NUMBER_L + (charcode-33)

   return ((colour << 13) | tilenumber)

# Return a list of 64 16-bit tegel values encoding the given input line.
def analyse_line(source):
   # By default all entries are a transparent tile zero.
   out = [0] * 64 
   
   xcursor = 0
   colour = 1
   
   # Affected by 0x06, 0x07 escapes for drawing double-height text.
   largetextmode = 0
   
   for c in source:
      code = c.encode("ascii")[0]
   
      if c >= ' ':
         out[xcursor] = tegel(code, colour, largetextmode)
         xcursor += 1
      elif code == 0x06:
         largetextmode = 0x06
      elif code == 0x07:
         largetextmode = 0x07
      elif code == 0x05:
         # The 0x05 escape inserts the current date/time.
         now = datetime.datetime.now()
         
         timestring = now.strftime("%Y/%m/%d %H:%M")
         
         for timechar in timestring:
            code = timechar.encode("ascii")[0]
            out[xcursor] = tegel(code, colour, largetextmode)
            xcursor += 1
      else:
         colour = code

   return out

def word_to_big_endian_bytes(word):
   return bytes((((word & 0xFF00) >> 8), (word & 0xFF)))

if __name__ == "__main__":
   parser = argparse.ArgumentParser()
   parser.add_argument("screen")
   parser.add_argument("output")
   args = parser.parse_args()
   
   with open(args.output,"wb") as o:
      current_screen = screens[args.screen]
      
      # Write the leading line count value first.
      o.write(word_to_big_endian_bytes(len(current_screen)))
      # Write the DBF loop counter variable.
      # This is the total number of tegels available in the block.
      o.write(word_to_big_endian_bytes(64*len(current_screen)-1))
   
      for line in (analyse_line(line) for line in current_screen):
         if len(line) != 64:
            raise Exception("Line "+str(line)+" is not 64 long.")
            
         for calculated_tegel in line:
            o.write(word_to_big_endian_bytes(calculated_tegel))
   